#include <SPI.h>
#include "BlockImage.h"
#include "BackgroundImage.h"
#include "Adafruit_GFX.h"
#include "ILI9341.h"

// Ili9341 LCD panel
#define TFT_DC          21
#define TFT_CS          D8
#define TFT_MOSI        MOSI
#define TFT_CLK         SCK
#define TFT_RST         22
#define TFT_MISO        MISO

//PS2 controller pin
#define VRX A0
#define VRY A1
#define SW  16

Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_RST);

uint16_t backBuffer[240][120];
const int Length = 12;     // the number of pixels for a side of a block
const int Width = 10;     // the number of horizontal blocks
const int Height = 20;    // the number of vertical blocks
int screen[Width][Height] = {0};// it shows color-numbers of all positions
uint16_t backcolor = 0;

struct Point {
  int X;
  int Y;
};

struct Block {
  Point square[4][4];
  int numRotate;
  int color;
};

Point pos;
Block block,NextBlock;

int rot,NextRot;
int fall_cnt = 0;
bool started = false;
bool gameover = false;
unsigned int score=0;

// Every block and rotate
Block blocks[7] = {
  {{{{ -1, 0}, {0, 0}, {1, 0}, {2, 0}}, {{0, -1}, {0, 0}, {0, 1}, {0, 2}}, {{0, 0}, {0, 0}, {0, 0}, {0, 0}}, {{0, 0}, {0, 0}, {0, 0}, {0, 0}}}, 2, 1},
  {{{{0, -1}, {1, -1}, {0, 0}, {1, 0}}, {{0, 0}, {0, 0}, {0, 0}, {0, 0}}, {{0, 0}, {0, 0}, {0, 0}, {0, 0}}, {{0, 0}, {0, 0}, {0, 0}, {0, 0}}}, 1, 2},
  {{{{ -1, -1}, { -1, 0}, {0, 0}, {1, 0}}, {{ -1, 1}, {0, 1}, {0, 0}, {0, -1}}, {{ -1, 0}, {0, 0}, {1, 0}, {1, 1}}, {{1, -1}, {0, -1}, {0, 0}, {0, 1}}}, 4, 3},
  {{{{ -1, 0}, {0, 0}, {0, 1}, {1, 1}}, {{0, -1}, {0, 0}, { -1, 0}, { -1, 1}}, {{0, 0}, {0, 0}, {0, 0}, {0, 0}}, {{0, 0}, {0, 0}, {0, 0}, {0, 0}}}, 2, 4},
  {{{{ -1, 0}, {0, 0}, {1, 0}, {1, -1}}, {{ -1, -1}, {0, -1}, {0, 0}, {0, 1}}, {{ -1, 1}, { -1, 0}, {0, 0}, {1, 0}}, {{0, -1}, {0, 0}, {0, 1}, {1, 1}}}, 4, 5},
  {{{{ -1, 1}, {0, 1}, {0, 0}, {1, 0}}, {{0, -1}, {0, 0}, {1, 0}, {1, 1}}, {{0, 0}, {0, 0}, {0, 0}, {0, 0}}, {{0, 0}, {0, 0}, {0, 0}, {0, 0}}}, 2, 6},
  {{{{ -1, 0}, {0, 0}, {1, 0}, {0, -1}}, {{0, -1}, {0, 0}, {0, 1}, { -1, 0}}, {{ -1, 0}, {0, 0}, {1, 0}, {0, 1}}, {{0, -1}, {0, 0}, {0, 1}, {1, 0}}}, 4, 7}
};

// Draw the screen
void Draw() { 
  static unsigned int lastscore=0xFFFF;
  static int LastBlockNumRotate=0xFFFF;
  static int LastBlockColor=0xFFFF;

  // Draw score and block only when they have changed
  if ((lastscore!=score)||(NextBlock.numRotate!=LastBlockNumRotate)||(NextBlock.color!=LastBlockColor)) { 
      // Read backg round left part to the buffer 
      for (int i = 0 ; i < 240; ++i) for (int j = 0; j < 120; ++j) {
         backBuffer[i][j]= BackgroundImage[i][j];}

      int putX=80,putY=20; 
      int x1,y1;
      for (int z=0;z<4;z++)
        for (int i = 0; i < Width; ++i) for (int j = 0; j < Height; ++j)
          for (int k = 0; k < Length; ++k) for (int l = 0; l < Length; ++l)
            {
                x1=putX+(NextBlock.square[NextRot][z].Y*Length+k);
                y1=putY+(NextBlock.square[NextRot][z].X*Length+l);
                backBuffer[x1][y1] = BlockImage[NextBlock.color][k][l];      
            }
      LastBlockNumRotate=NextBlock.numRotate;
      LastBlockColor=NextBlock.color;

      //Convert backBuffer to 100*240
      for (int i = 0 ; i < 240; ++i) for (int j = 0; j < 100; ++j) {
         backBuffer[(i*100+j)/120][(i*100+j)%120]= backBuffer[i][j];}
      
      //Draw buffer
      tft.fillImage(backBuffer, 0, 0, 100, 240);
      
      // Draw score to the screen
      tft.setCursor(10, 120);
      tft.setTextColor(ILI9341_OLIVE);  
      tft.setTextSize(2);
      char tmp[20]; 
      sprintf(tmp, "%04d", score);
      tft.println(tmp);
      lastscore=score;
  }
  
  // Draw blocks in the main area   
  for (int i = 0; i < Width; ++i) for (int j = 0; j < Height; ++j)
      for (int k = 0; k < Length; ++k) for (int l = 0; l < Length; ++l)
          backBuffer[j * Length + l][i * Length + k] = BlockImage[screen[i][j]][k][l];
  tft.fillImage(backBuffer, 100, 0, 120, 240);
}

// Put a new block in start position
void PutStartPos() {
  pos.X = 4;
  pos.Y = 1;
  block=NextBlock;
  rot=NextRot;
  int c=random(7);
  NextBlock = blocks[c];
  NextRot = random(NextBlock.numRotate);  
}

bool GetSquares(Block block, Point pos, int rot, Point* squares) {
  bool overlap = false;
  for (int i = 0; i < 4; ++i) {
    Point p;
    p.X = pos.X + block.square[rot][i].X;
    p.Y = pos.Y + block.square[rot][i].Y;
    overlap |= p.X < 0 || p.X >= Width || p.Y < 0 || p.Y >= Height || screen[p.X][p.Y] != 0;
    squares[i] = p;
  }
  return !overlap;
}

// Test if it's game over now
void GameOver() {
  for (int i = 0; i < Width; ++i) for (int j = 0; j < Height; ++j) if (screen[i][j] != 0) screen[i][j] = 4;
  gameover = true;
}

boolean Button_A = false;
boolean Button_UP = false;
boolean Button_DOWN = false;
boolean Button_LEFT = false;
boolean Button_RIGHT = false;

void ClearKeys() {
  Button_A=false;
  Button_UP=false;
  Button_DOWN=false;
  Button_LEFT=false;
  Button_RIGHT=false;
}

unsigned long ButtonElsp=0;
bool KeyPadLoop(){
  if (millis()-ButtonElsp<100) {return false;}
  ButtonElsp=millis();
  ClearKeys();
  if(analogRead(VRX)<1000) {
        Button_LEFT=true;
        Serial.println("left"); 
        return true; }
  if(analogRead(VRX)>3000) {
        Button_RIGHT=true;
        Serial.println("right"); 
        return true;}
  if(analogRead(VRY)<300) {
        Button_UP=true;
        Serial.print("up  "); 
        Serial.println(ButtonElsp); 
        return true;}
  if(analogRead(VRY)>3000) {
        Button_DOWN=true;
        Serial.println("down"); 
        return true;}
  if(digitalRead(SW)==0) {
        Button_A=true;
        Serial.println("start"); 
        return true;}
  return false;
}

/*
<10 0  500  500-(score/10)*100
<20 1  400
<30  300
<40  200
<50  100
>=50  50
 */
unsigned long Score2Level ()
{ 
  if (score>=200) {return 50;}
  return (2000-(score/10)*100);
}
unsigned long int TimeElsp;
void GetNextPosRot(Point* pnext_pos, int* pnext_rot) {
  bool received = KeyPadLoop();
  if (Button_A) started = true;
  if (!started) return;
  pnext_pos->X = pos.X;
  pnext_pos->Y = pos.Y;
  //if ((fall_cnt = (fall_cnt + 1) % 10) == 0) pnext_pos->Y += 1;
  if ((millis()-TimeElsp)>Score2Level()) {pnext_pos->Y += 1;TimeElsp=millis();}
  else if (received) {
    if (Button_LEFT) {
      Button_LEFT = false;
      pnext_pos->X -= 1;
    }
    else if (Button_RIGHT) {
      Button_RIGHT = false;
      pnext_pos->X += 1;
    }
    else if (Button_DOWN) {
      Button_DOWN = false;
      pnext_pos->Y += 1;
    }
    else if (Button_UP) {
      Button_A = false;
      *pnext_rot = (*pnext_rot + block.numRotate - 1)%block.numRotate; 
    }
  }
}

void DeleteLine() {
  int DeleteCount=0;
  for (int j = 0; j < Height; ++j) {
    bool Delete = true;
    for (int i = 0; i < Width; ++i) if (screen[i][j] == 0) Delete = false; 
    if (Delete) { 
      for (int k = j; k >= 1; --k) for (int i = 0; i < Width; ++i) screen[i][k] = screen[i][k - 1];
      DeleteCount++; 
    }
  }
  if (DeleteCount!=0) {score+=DeleteCount*DeleteCount;}
}

void ReviseScreen(Point next_pos, int next_rot) {
  if (!started) return;
  Point next_squares[4];
  for (int i = 0; i < 4; ++i) screen[pos.X + block.square[rot][i].X][pos.Y + block.square[rot][i].Y] = 0;
  if (GetSquares(block, next_pos, next_rot, next_squares)) {
    for (int i = 0; i < 4; ++i) screen[next_squares[i].X][next_squares[i].Y] = block.color;
    pos = next_pos;
    rot = next_rot;
  }
  else {
    for (int i = 0; i < 4; ++i) screen[pos.X + block.square[rot][i].X][pos.Y + block.square[rot][i].Y] = block.color;
    if (next_pos.Y == pos.Y + 1) {
      DeleteLine();
      PutStartPos();
      if (!GetSquares(block, pos, rot, next_squares)) {
        for (int i = 0; i < 4; ++i) screen[pos.X + block.square[rot][i].X][pos.Y + block.square[rot][i].Y] = block.color;
        GameOver();
      }
    }
  }
  Draw();
}

void InitScreen() {
  NextBlock = blocks[random(7)];
  NextRot = random(NextBlock.numRotate);
  
  tft.fillScreen(backcolor);
  for (int i = 0 ; i < 240; ++i) for (int j = 0; j < 120; ++j) backBuffer[i][j] = BackgroundImage[i][j];
  tft.fillImage(backBuffer, 0, 0, 120, 240);
  for (int i = 0 ; i < 240; ++i) for (int j = 0; j < 120; ++j) backBuffer[i][j] = BackgroundImage[i][j + 200];
  tft.fillImage(backBuffer, 200, 0, 120, 240);

  
  PutStartPos();
  memset(screen,0,sizeof(screen));
  for (int i = 0; i < 4; ++i) screen[pos.X + block.square[rot][i].X][pos.Y + block.square[rot][i].Y] = block.color;
  Draw(); 
}

void setup(void) {
  Serial.begin(115200);

  pinMode(VRX,INPUT);
  pinMode(VRY,INPUT);
  pinMode(SW,INPUT_PULLUP);

  tft.begin(20000000UL);
  tft.setRotation(1);
    
  InitScreen();
}
void loop() {
  while (gameover==true) {
      if ((KeyPadLoop())&&(Button_A)) {
          InitScreen();
          gameover=false;
          started=false;
          score=0;
        }
      delay(300);
    } 
  Point next_pos;
  int next_rot = rot;
  GetNextPosRot(&next_pos, &next_rot);
  ReviseScreen(next_pos, next_rot);
}
